# -*- coding: utf-8 -*-

import pylab

#直線の幅を指定する
pylab.rcParams['lines.linewidth'] = 4
#タイトルのフォントサイズを指定する
pylab.rcParams['axes.titlesize'] = 20
#軸のラベルのフォントサイズを指定する
pylab.rcParams['axes.labelsize'] = 20
#x軸上の数値のサイズを指定する
pylab.rcParams['xtick.labelsize'] = 16
#y軸上の数値のサイズを指定する
pylab.rcParams['ytick.labelsize'] = 16
#x軸の目盛のサイズを指定する
pylab.rcParams['xtick.major.size'] = 7
#y軸の目盛のサイズを指定する
pylab.rcParams['ytick.major.size'] = 7
#マーカー（例えば点を表す円）のサイズを指定する．
#凡例のnumpointsを指定する．
pylab.rcParams['legend.numpoints'] = 1

#figure 18.1 contains no code

#Figure 18.2
def getData(fileName):
    dataFile = open(fileName, 'r')
    distances = []
    masses = []
    dataFile.readline() #ヘッダーは無視する
    for line in dataFile:
        d, m = line.split(' ')
        distances.append(float(d))
        masses.append(float(m))
    dataFile.close()
    return (masses, distances)

#Figure 18.3
def plotData(inputFile):
    masses, distances = getData(inputFile)
    distances = pylab.array(distances) 
    masses = pylab.array(masses)
    forces = masses*9.81
    pylab.plot(forces, distances, 'bo',
               label = 'Measured displacements')
    pylab.title('Measured Displacement of Spring')
    pylab.xlabel('|Force| (Newtons)')
    pylab.ylabel('Distance (meters)')
    
plotData('springData.txt')

#Figure 18.4 has no code

#Figure 18.5
def fitData(inputFile):
    masses, distances = getData(inputFile)
    distances = pylab.array(distances)
    forces = pylab.array(masses)*9.81
    pylab.plot(forces, distances, 'ko',
               label = 'Measured displacements')
    pylab.title('Measured Displacement of Spring')
    pylab.xlabel('|Force| (Newtons)')
    pylab.ylabel('Distance (meters)')
    #線形回帰を行って係数を求める．
    a,b = pylab.polyfit(forces, distances, 1)
    predictedDistances = a*pylab.array(forces) + b
    k = 1.0/a #テキスト中の説明を参照せよ
    pylab.plot(forces, predictedDistances,
              label = 'Displacements predicted by\nlinear fit, k = '
               + str(round(k, 5)))
    pylab.legend(loc = 'best')

#Figure 18.6-18.10 have no code

#Figure 18.11
def getTrajectoryData(fileName):
    dataFile = open(fileName, 'r')
    distances = []
    heights1, heights2, heights3, heights4 = [],[],[],[]
    dataFile.readline()
    for line in dataFile:
        d, h1, h2, h3, h4 = line.split()
        distances.append(float(d))
        heights1.append(float(h1))
        heights2.append(float(h2))
        heights3.append(float(h3))
        heights4.append(float(h4))
    dataFile.close()
    return (distances, [heights1, heights2, heights3, heights4])

def processTrajectories(fileName):
    distances, heights = getTrajectoryData(fileName)
    numTrials = len(heights)
    distances = pylab.array(distances)
    #各距離での平均高さを要素とする配列を得る．
    totHeights = pylab.array([0]*len(distances))
    for h in heights:
        totHeights = totHeights + pylab.array(h)
    meanHeights = totHeights/len(heights)
    pylab.title('Trajectory of Projectile (Mean of '\
                + str(numTrials) + ' Trials)')
    pylab.xlabel('Inches from Launch Point')
    pylab.ylabel('Inches Above Launch Point')
    pylab.plot(distances, meanHeights, 'ko')
    fit = pylab.polyfit(distances, meanHeights, 1)
    altitudes = pylab.polyval(fit, distances)
    pylab.plot(distances, altitudes, 'b', label = 'Linear Fit')
    fit = pylab.polyfit(distances, meanHeights, 2)
    altitudes = pylab.polyval(fit, distances)
    pylab.plot(distances, altitudes, 'k:', label = 'Quadratic Fit')
    pylab.legend()

processTrajectories('launcherData.txt')

#Figure 18.12 has not code

#Figure 18.13
def rSquared(measured, predicted):
    """measuredは計測値を要素とする1次元配列，predictedは予測値を要素とする1次元配列と仮定する．
       この関数は決定係数を返す． """
    estimateError = ((predicted - measured)**2).sum()
    meanOfMeasured = measured.sum()/len(measured)
    variability = ((measured - meanOfMeasured)**2).sum()
    return 1 - estimateError/variability
    
#Figure 18.14
def getHorizontalSpeed(quadFit, minX, maxX):
    """quadFitは2次の多項式の係数と仮定する．minXとmaxXは距離とする（単位はインチ）．
        この関数はフィート毎秒を単位とする水平方向の速度を返す．
        """
    inchesPerFoot = 12
    xMid = (maxX - minX)/2
    a,b,c = quadFit[0], quadFit[1], quadFit[2]
    yPeak = a*xMid**2 + b*xMid + c
    g = 32.16*inchesPerFoot #重力加速度（単位はインチ毎秒毎秒）
    t = (2*yPeak/g)**0.5 #頂点から目的点までの所要時間（単位は秒）
    print('Horizontal speed =',
          int(xMid/(t*inchesPerFoot)), 'feet/sec')

#Figure 18.15
pylab.figure()
vals = []
for i in range(10):
    vals.append(3**i)
pylab.plot(vals,'ko', label = 'Actual points')
xVals = pylab.arange(10)
fit = pylab.polyfit(xVals, vals, 5)
yVals = pylab.polyval(fit, xVals)
pylab.plot(yVals, 'kx', label = 'Predicted points',
           markeredgewidth = 2, markersize = 25)
pylab.title('Fitting y = 3**x')
pylab.legend(loc = 'upper left')

#Figures 18.15 - `8.17 contain no code

#figure 18.18
import math

def createData(f, xVals):
    """fは1変数関数，xValsは関数fの変数として適切な値を要素とする配列と仮定する．
        この関数は，fをxValsの各要素に適用した結果を含む配列を返す．"""
    yVals = []
    for i in xVals:
        yVals.append(f(xVals[i]))
    return pylab.array(yVals)

def fitExpData(xVals, yVals):
    """Assumes xVals and yVals arrays of numbers such that
         yVals[i] == f(xVals[i]), where f is an exponential function
       Returns a, b, base such that log(f(x), base) == ax + b"""
    """xValsとyValsは，yVals[i]==f(xVals[i])を満たす値の配列と仮定する．ただし，fは指数関数とする．
        この関数は，log(f(x),base)==ax+bを満たすa,b,baseを返す．"""
    logVals = []
    for y in yVals:
        logVals.append(math.log(y, 2.0)) #底を2とする対数
    fit = pylab.polyfit(xVals, logVals, 1)
    return fit, 2.0

xVals = range(10)
f = lambda x: 3**x
yVals = createData(f, xVals)
pylab.plot(xVals, yVals, 'ko', label = 'Actual values')
fit, base = fitExpData(xVals, yVals)
predictedYVals = []
for x in xVals:
    predictedYVals.append(base**pylab.polyval(fit, x))
pylab.plot(xVals, predictedYVals, label = 'Predicted values')
pylab.title('Fitting an Exponential Function')
pylab.legend(loc = 'upper left')
#オリジナルなデータには含まれていないxに対する値を探す
print('f(20) =', f(20))
print('Predicted value =', int(base**(pylab.polyval(fit, [20]))))

